# pylint: disable=duplicate-code
"""
The module to check the Druid retention rule
"""
import argparse
import base64
import json
import logging
import re
import ssl
import time
from urllib.error import HTTPError
from urllib.error import URLError
from urllib.parse import urlparse
from urllib.request import Request
from urllib.request import urlopen
from tasks.task_common import read_json_file


FORMAT = '%(asctime)s %(levelname)s: %(message)s'
DATEFMT = '%Y-%m-%d %H:%M:%S'
TIMEOUT_LOWER_BOUND = 15
TIMEOUT_HIGHER_BOUND = 30
DEFAULT_RULE = '_default'
logging.basicConfig(format=FORMAT, datefmt=DATEFMT, level=logging.DEBUG)


def raise_friendly_error(exception):
    """
    Raise friendly error
    """
    if isinstance(exception, HTTPError):
        text = exception.read().decode('utf-8').strip()
        result = re.search(r"<pre>(.*?)</pre>", text, re.DOTALL)
        if result:
            text = result.group(1).strip()
        raise Exception(f"HTTP Error {exception.code}: {exception.reason}, " +
                        f"check coordinator log for more details.\n{text}")
    raise exception


def add_basic_auth_header(args, req):
    """
    Add basic auth header
    """
    if args.user is not None:
        basic_auth_encoded = base64.b64encode(f"{args.user}:{args.password}")
        req.add_header("Authorization", f"Basic {basic_auth_encoded}")


def get_retention_rules(args, timeout_at, ctx=None):
    """
    Get retention rules
    """
    url = args.coordinator_url.rstrip("/") + "/druid/coordinator/v1/rules/"
    req = Request(url)
    add_basic_auth_header(args, req)
    timeleft = timeout_at - time.time()
    response_timeout = min(max(timeleft, TIMEOUT_LOWER_BOUND), TIMEOUT_HIGHER_BOUND)
    with urlopen(req, None, response_timeout, context=ctx) as response:
        response_obj = json.loads(response.read())
        return response_obj


def delete_rule(args, datasource, timeout_at, ctx=None):
    # delete the rule by posting empty rule
    post_rule(args, datasource, "[]", timeout_at, ctx)


def open_url_with_retry(request, data, max_retries=3, timeout=10, delay_between_retries=10, ctx=None):
    for attempt in range(max_retries):
        try:
            with urlopen(request, data, timeout=timeout, context=ctx) as response:
                return response.msg
        except URLError as e:
            logging.warn(f"Attempt {attempt + 1} failed: {e.reason}")
            if attempt < max_retries - 1:
                time.sleep(delay_between_retries)
            else:
                raise


def post_rule(args, datasource, rule_json, timeout_at, ctx=None):
    """
    Post retention rules
    """
    try:
        url = args.coordinator_url.rstrip(
            "/") + "/druid/coordinator/v1/rules/" + datasource
        data = rule_json.encode("utf-8")
        req = Request(
            url, None, {'Content-Type': 'application/json', 'Accept': 'application/json'})
        add_basic_auth_header(args, req)
        timeleft = timeout_at - time.time()
        response_timeout = min(max(timeleft, TIMEOUT_LOWER_BOUND), TIMEOUT_HIGHER_BOUND)
        response = open_url_with_retry(req, data, 5, response_timeout, 10, ctx)
        logging.info("Post retention rule response %s", response)
        return None

    except URLError as error:
        # pylint: disable=E1101
        if isinstance(error, HTTPError) and error.code >= 400 and error.code <= 500:
            # 4xx (problem with the request) or 500 (something wrong on the server)
            raise_friendly_error(error)
        elif time.time() >= timeout_at:
            # No futher retries
            logging.info("Timed out posting rules.")
            raise_friendly_error(error)
        elif isinstance(error, HTTPError) and error.code in [301, 302, 303, 305, 307] and \
                error.getheader("Location") is not None:
            # Set the new location in args.coordinator_url
            # so it can be used by post_rule to re-issue request
            location = urlparse(error.getheader("Location"))
            args.coordinator_url = f"{location.scheme}://{location.netloc}"
            logging.info("Redirect response received, setting url to %s", args.url)
            return post_rule(args, rule_json, timeout_at, ctx)
        else:
            # If at first you don't succeed, try, try again!
            sleep_time = 5
            logging.info("Retrying...")
            time.sleep(sleep_time)
            return post_rule(args, rule_json, timeout_at, ctx)
        return None


def main():
    """
    Main function
    """
    parser = argparse.ArgumentParser(
        description='Check Druid retention rules.')
    parser.add_argument('--file', '-f', type=str,
                        required=True, help='Retention rules JSON file')
    parser.add_argument('--coordinator-url', type=str,
                        default='https://druid-coordinator:8281/', help='Druid Coordinator url')
    parser.add_argument('--timeout', type=int, default=120,
                        help='Timeout (in seconds) for posting rules')
    parser.add_argument('--user', type=str, default=None,
                        help='Basic auth username')
    parser.add_argument('--password', type=str,
                        default=None, help='Basic auth password')
    args = parser.parse_args()

    # SSL context initialization
    ctx = ssl.create_default_context()
    ctx.load_verify_locations("/etc/nsxi/certs/ca.crt")

    timeout_at = time.time() + args.timeout
    target_retention_rules = json.loads(read_json_file(args.file))
    logging.info("Target retention rules: " + str(target_retention_rules))
    current_rules = get_retention_rules(args, timeout_at, ctx)
    logging.info("Current retention rules:" + str(current_rules))
    # delete all the other rules not in targetRule, except the default rule
    for datasource in current_rules.keys():
        if datasource == DEFAULT_RULE:
            continue
        if datasource not in target_retention_rules and len(current_rules[datasource]) > 0:
            logging.info("delete rule for datasource: " + datasource)
            timeout_at = time.time() + args.timeout
            delete_rule(args, datasource, timeout_at, ctx)
            time.sleep(2)

    # add target retention rule to current rule if not exist or empty
    for datasource in target_retention_rules.keys():
        if datasource not in current_rules or not current_rules[datasource]:
            logging.info("Rule for datasource: " + datasource +
                         " does not exist, adding the rule: " + str(target_retention_rules.get(datasource)))
            timeout_at = time.time() + args.timeout
            post_rule(args, datasource, json.dumps(target_retention_rules[datasource]), timeout_at, ctx)
            time.sleep(2)


if __name__ == "__main__":
    main()
