#!/usr/bin/env python
"""
The module to run compaction task
"""
import argparse
import json
import logging
import ssl
import sys
import time
from datetime import datetime, timedelta

from tasks.task_common import await_load_completion
from tasks.task_common import await_task_completion
from tasks.task_common import post_task

FORMAT = '%(asctime)s %(levelname)s: %(message)s'
DATEFMT = '%Y-%m-%d %H:%M:%S'
logging.basicConfig(format=FORMAT, datefmt=DATEFMT, level=logging.DEBUG)

DEFAULT_COMPACTION_SPEC = {
    "type": "compact",
    "dataSource": "",
    "ioConfig": {
        "type": "compact",
        "inputSpec": {
            "type": "interval",
            "interval": ""
        }
    }
}


def get_time_interval(args):
    """
    Get time interval
    """
    if args.start_timestamp is not None and args.end_timestamp is not None:
        interval_start_timestamp = datetime.strptime(
            args.start_timestamp, '%Y-%m-%dT%H:00:00')
        interval_end_timestamp = datetime.strptime(
            args.end_timestamp, '%Y-%m-%dT%H:00:00')
        time_interval = f"{interval_start_timestamp}/{interval_end_timestamp}"
        return time_interval

    days = 1 if args.num_days is None else args.num_days
    end_date = datetime.today().strftime('%Y-%m-%d')
    start_date = (datetime.today() - timedelta(days)).strftime('%Y-%m-%d')
    time_interval = f"{start_date}/{end_date}"
    return time_interval


def main():
    """
    Main function to run compaction task
    """
    parser = argparse.ArgumentParser(
        description='Post Druid compaction tasks.')
    parser.add_argument('--data-source', '-d', type=str, required=True,
                        default=None, help='The name of Druid table (dataSource) for compaction')
    parser.add_argument('--start-timestamp', '-start', type=str, default=None,
                        help='Starting timestamp from which compaction should happen')
    parser.add_argument('--end-timestamp', '-end', type=str, default=None,
                        help='Ending timestamp from which compaction should happen')
    parser.add_argument('--num-days', '-n', type=int, default=None,
                        help='Number of days from when the compaction should happen')
    parser.add_argument('--url', '-u', metavar='url', type=str,
                        default='https://druid-coordinator:8281/', help='Druid Overlord url')
    parser.add_argument('--coordinator-url', type=str,
                        default='https://druid-coordinator:8281/', help='Druid Coordinator url')
    parser.add_argument('--submit-timeout', type=int, default=120,
                        help='Timeout (in seconds) for submitting tasks')
    parser.add_argument('--complete-timeout', type=int, default=14400,
                        help='Timeout (in seconds) for completing tasks')
    parser.add_argument('--load-timeout', type=int, default=14400,
                        help='Timeout (in seconds) for waiting for tasks to load')
    parser.add_argument('--quiet', '-q', action='store_true',
                        help='Suppress retryable errors')
    parser.add_argument('--user', type=str, default=None,
                        help='Basic auth username')
    parser.add_argument('--password', type=str,
                        default=None, help='Basic auth password')
    args = parser.parse_args()

    submit_timeout_at = time.time() + args.submit_timeout
    complete_timeout_at = time.time() + args.complete_timeout

    task_spec = DEFAULT_COMPACTION_SPEC
    task_spec["dataSource"] = args.data_source
    task_spec["ioConfig"]["inputSpec"]["interval"] = get_time_interval(args)

    task_contents = json.dumps(task_spec)

    print("task_spec:", flush=True)
    print(task_contents, flush=True)

    # SSL context initialization
    ctx = ssl.create_default_context()
    ctx.load_verify_locations("/etc/nsxi/certs/ca.crt")

    task_json = json.loads(task_contents)
    if task_json['type'] == "compact":
        datasource = task_json['dataSource']
    else:
        datasource = json.loads(task_contents)[
            "spec"]["dataSchema"]["dataSource"]

    logging.info("Beginning indexing data for %s\n", datasource)
    sys.stdout.flush()
    task_id = json.loads(post_task(args, task_contents,
                                   submit_timeout_at, ctx))["task"]

    # pylint: disable=W1201,C0209
    logging.info('\033[1m' + "Task started: " + '\033[0m' +
                 "{0}".format(task_id))
    logging.info('\033[1m' + "Task log:     " + '\033[0m' +
                 "{0}/druid/indexer/v1/task/{1}/log".format(args.url.rstrip("/"), task_id))
    logging.info('\033[1m' + "Task status:  " + '\033[0m' +
                 "{0}/druid/indexer/v1/task/{1}/status".format(args.url.rstrip("/"), task_id))

    task_status = await_task_completion(
        args, task_id, complete_timeout_at, ctx)
    logging.info("Task finished with status: %s", task_status)
    if task_status != 'SUCCESS':
        sys.exit(1)

    logging.info("Completed indexing data for %s. "
                 "Now loading indexed data onto the cluster...", datasource)
    load_timeout_at = time.time() + args.load_timeout
    await_load_completion(args, datasource, load_timeout_at, ctx)


if __name__ == "__main__":
    main()
