import argparse
import logging
import os

from kubernetes import client, config
from kubernetes.client.rest import ApiException

BUILD_TAG = os.environ.get("BUILD_TAG")
FORMAT = '%(asctime)s %(levelname)s: %(message)s'
DATEFMT = '%Y-%m-%d %H:%M:%S'
DRUID_HISTORICAL_LABEL = "component=historical"
DRUID_CONFIG_HISTORICAL_LABEL = "component=config-historical"
logging.basicConfig(format=FORMAT, datefmt=DATEFMT, level=logging.INFO)


def delete_historical_pvc(ns, label, app_v1, core_v1):
    historical_sts = app_v1.read_namespaced_stateful_set(
        namespace=ns, name="druid-historical")
    historical_replicas = historical_sts.spec.replicas

    historical_pvc_list = core_v1.list_namespaced_persistent_volume_claim(namespace=ns,
                                                                          label_selector=label)

    # When the historical replica after upgrade is less than historical pvc number, delete stale pvc
    if len(historical_pvc_list.items) > historical_replicas:
        logging.info(
            f"It has {len(historical_pvc_list.items) - historical_replicas} stale pvc for historical")
        for pvc in historical_pvc_list.items:
            split_pvc_name = pvc.metadata.name.split("-")
            index = split_pvc_name[-1]
            if (index.isdigit()) and int(index) > historical_replicas - 1:
                logging.info(f"Deleting stale pvc {pvc.metadata.name}")
                response = core_v1.delete_namespaced_persistent_volume_claim(name=pvc.metadata.name, namespace=ns,
                                                                             body=client.V1DeleteOptions(
                                                                                 propagation_policy="Background"))
    else:
        logging.info(f"No stale pvc for Druid historical")


def delete_config_historical_pvc(ns, label, app_v1, core_v1):
    stateful_sets = app_v1.list_namespaced_stateful_set(ns)
    stateful_set_exists = any(stateful_set.metadata.name ==
                              "druid-config-historical" for stateful_set in stateful_sets.items)

    # Delete config-historical pvc when config historical sts not exist
    if not stateful_set_exists:
        logging.info(
            f"StatefulSet druid-config-historical not exist. Deleting stale pvc has label {label}")
        response = core_v1.delete_collection_namespaced_persistent_volume_claim(label_selector=label, namespace=ns,
                                                                                body=client.V1DeleteOptions(
                                                                                    propagation_policy="Background"))
    else:
        logging.info(f"StatefulSet that has label {label} exists. Do nothing")


def main():
    parser = argparse.ArgumentParser(description='Redis pvc upgrade.')
    parser.add_argument('--name-space', '-ns', type=str,
                        default='nsxi-platform', help='Name space of the sts')
    parser.add_argument('--labels', '-l', metavar='STRING', nargs='+', required=True,
                        help='The label for the sts/pvc')
    args = parser.parse_args()

    try:
        if BUILD_TAG.startswith("kindtest"):
            logging.info("Skipping druid delete pvc job for kind test")
            return
        labels = args.labels

        config.load_incluster_config()
        app_v1 = client.AppsV1Api()
        core_v1 = client.CoreV1Api()
        ns = args.name_space

        for label in labels:
            if label == DRUID_CONFIG_HISTORICAL_LABEL:
                delete_config_historical_pvc(ns, label, app_v1, core_v1)
            elif label == DRUID_HISTORICAL_LABEL:
                delete_historical_pvc(ns, label, app_v1, core_v1)

    except ApiException as e:
        logging.info(f"Error deleting pvc: {e}")


if __name__ == "__main__":
    main()
