import argparse
import copy
import logging
import os
import re
import sys
import time

from kubernetes import client, config
from kubernetes.client.rest import ApiException

MULTIPLIER_DICT = {"K": 1000, "Ki": 1024, "M": 1000000, "Mi": 1048576, "G": 1000000000, "Gi": 1073741824}

BUILD_TAG = os.environ.get("BUILD_TAG")
FORMAT = '%(asctime)s %(levelname)s: %(message)s'
DATEFMT = '%Y-%m-%d %H:%M:%S'
logging.basicConfig(format=FORMAT, datefmt=DATEFMT, level=logging.INFO)


def update_druid_historical_replica(app_v1, ns, new_replica):
    historical_sts = app_v1.read_namespaced_stateful_set(namespace=ns, name="druid-historical")

    old_historical_replica = copy.deepcopy(historical_sts.spec.replicas)
    historical_sts.spec.replicas = new_replica
    app_v1.patch_namespaced_stateful_set_scale(
        name="druid-historical",
        namespace=ns,
        body=historical_sts
    )
    return old_historical_replica


def update_druid_historical_pvc_size(core_v1, ns, pvc_name, size):
    gi_size = str(size) + "Gi"
    try:
        patch = [
            {"op": "replace", "path": "/spec/resources/requests/storage", "value": gi_size}
        ]
        core_v1.patch_namespaced_persistent_volume_claim(namespace=ns, name=pvc_name, body=patch)
    except ApiException as e:
        logging.info(f"Error patch pvc: {e}")
        sys.exit("Failed to patch pvc")


def calculate_pvc_bytes_size(pvc):
    pvc_name = pvc.metadata.name
    storage = pvc.spec.resources.requests["storage"]
    capacity = re.findall(r'[A-Za-z]+|\d+', storage)
    if len(capacity) == 2:
        size = capacity[0]
        unit = capacity[1]
        if unit not in MULTIPLIER_DICT:
            sys.exit("Failed to parse the pvc size")
        size_bytes = int(size) * MULTIPLIER_DICT[unit]
        logging.info(f"{capacity} = byte {size_bytes}")
    elif len(capacity) == 1:
        size_bytes = capacity[0]
    else:
        sys.exit("Failed to parse the pvc size")

    return int(size_bytes)


def get_storage_class_allow_volume_expansion_for_sts(sts_name, ns, app_v1, storage_v1):
    try:
        historical_sts = app_v1.read_namespaced_stateful_set(namespace=ns, name=sts_name)
    except ApiException as e:
        if e.status == 404:
            print(f"StatefulSet '{sts_name}' not found. Exiting gracefully.")
            sys.exit(0)
    storage_class_name = historical_sts.spec.volume_claim_templates[0].spec.storage_class_name
    storage_class = storage_v1.read_storage_class(name=storage_class_name)
    if storage_class and hasattr(storage_class, "allow_volume_expansion"):
        allow_volume_expansion = storage_class.allow_volume_expansion
        logging.info(f"StorageClass: {storage_class.metadata.name}")
        logging.info(f"allowVolumeExpansion: {allow_volume_expansion}")
        return allow_volume_expansion
    else:
        logging.info(f"StorageClass {storage_class_name} not found or does not support volume expansion.")
        return False


def resize_pvc_for_sts(sts_name, ns, label, app_v1, core_v1, size):
    historical_pvc_list = core_v1.list_namespaced_persistent_volume_claim(namespace=ns,
                                                                          label_selector=label)
    need_restart_sts = False
    for pvc in historical_pvc_list.items:
        size_bytes = calculate_pvc_bytes_size(pvc)
        target_size_bytes = size * 1073741824
        if size_bytes >= target_size_bytes:
            logging.info(f"No need to resize the pvc as current size {size_bytes} is larger than {size}Gi")
        else:
            update_druid_historical_pvc_size(core_v1, ns, pvc.metadata.name, size)
            need_restart_sts = True

    time.sleep(10)

    if need_restart_sts:
        logging.info("Changing the sts replica to 0")
        historical_replica = update_druid_historical_replica(app_v1, ns, 0)
        time.sleep(10)
        logging.info(f"Deleting the old sts and new sts will be created during helm upgrade")
        app_v1.delete_namespaced_stateful_set(name=sts_name, namespace=ns,
                                              body=client.V1DeleteOptions(propagation_policy="Foreground"))
    else:
        logging.info("No need to restart sts")


def delete_sts_and_pvc(sts_name, ns, label, app_v1, core_v1):
    app_v1.delete_namespaced_stateful_set(name=sts_name, namespace=ns,
                                          body=client.V1DeleteOptions(propagation_policy="Foreground"))

    historical_pvc_list = core_v1.list_namespaced_persistent_volume_claim(namespace=ns,
                                                                          label_selector=label)
    for pvc in historical_pvc_list.items:
        core_v1.delete_namespaced_persistent_volume_claim(name=pvc.metadata.name, namespace=ns,
                                                          body=client.V1DeleteOptions(
                                                              propagation_policy="Foreground"))


def main():
    parser = argparse.ArgumentParser(description='Druid pvc upgrade.')
    parser.add_argument('--stateful-set', '-sts', type=str, required=True,
                        help='The name of stateful set to be upgraded')
    parser.add_argument('--label', '-l', type=str, required=True,
                        help='The label for the sts')
    parser.add_argument('--name-space', '-ns', type=str,
                        default='nsxi-platform', help='Name space of the sts')
    parser.add_argument('--size', '-s', type=int, default=128,
                        required=True, help='New pvc Gi size')
    args = parser.parse_args()

    try:
        if BUILD_TAG.startswith("kindtest"):
            logging.info("Skipping pvc upgrade for kind test")
            return
        stateful_set_name = args.stateful_set
        label = args.label

        config.load_incluster_config()
        app_v1 = client.AppsV1Api()
        core_v1 = client.CoreV1Api()
        storage_v1 = client.StorageV1Api()
        ns = args.name_space

        allow_volume_expansion_enabled = get_storage_class_allow_volume_expansion_for_sts(stateful_set_name, ns, app_v1,
                                                                                          storage_v1)

        # If storage class allow volume expansion, it will edit pvc size and restart the stateful set.
        # If not allow, it will delete sts and pvc. During helm upgrade, sts and pvc with new size will be created.
        if allow_volume_expansion_enabled:
            logging.info("Allow volume expansion, will resize pvc")
            resize_pvc_for_sts(stateful_set_name, ns, label, app_v1, core_v1, args.size)
        else:
            logging.info("Not allow volume expansion, will delete pvc and sts")
            delete_sts_and_pvc(stateful_set_name, ns, label, app_v1, core_v1)

    except ApiException as e:
        logging.info(f"Error upgrading pvc: {e}")


if __name__ == "__main__":
    main()
