"""
The module is for submitting Druid reindex job when sites offboarded
"""
import argparse
import datetime
import json
import logging
import ssl
import sys
import time
from datetime import datetime, timedelta

from tasks.config import Config
from tasks.task_common import get_s2_sites
from tasks.task_common import post_query
from tasks.task_common import post_task
from tasks.task_common import read_json_file

FORMAT = '%(asctime)s %(levelname)s: %(message)s'
DATEFMT = '%Y-%m-%d %H:%M:%S'
PROCESSING_BUFFER_SIZE = 'PROCESSING_BUFFER_SIZE'
JAVA_OPTS = 'JAVA_OPTS'
logging.basicConfig(format=FORMAT, datefmt=DATEFMT, level=logging.DEBUG)


def get_offboarded_site_ids(args, ctx):
    """
    Get offboarded site ids from site services
    """
    submit_timeout_at = time.time() + args.submit_timeout
    offboarded_site_ids = set()
    response = get_s2_sites(args, submit_timeout_at, ctx)
    json_response = json.loads(response)
    sites_info = json_response['sites_info']
    for site_info in sites_info:
        if site_info['status'] == 'OFFBOARDED':
            offboarded_site_ids.add(site_info['site_id'])

    return offboarded_site_ids


def get_unique_site_ids_from_druid(data_source, dimensions, args, ctx):
    """
    Get existing site ids from Druid flow table
    """
    submit_timeout_at = time.time() + args.submit_timeout
    site_ids = set()
    # TopN version
    for dimension in dimensions:
        json_body = {
            "dataSource": data_source,
            "dimension": dimension,
            "granularity": "all",
            "intervals": [
                "1900-01-09T00:00:00.000Z/2992-01-10T00:00:00.000Z"
            ],
            "metric": {
                "type": "dimension",
                "ordering": "lexicographic"
            },
            "queryType": "topN",
            "threshold": 1000
        }
        response = post_query(args, json.dumps(
            json_body), submit_timeout_at, ctx)

        json_response = json.loads(response)
        for topn_response in json_response:
            for element in topn_response['result']:
                site_ids.add(element[dimension])

    return site_ids


def get_exclude_filter(reindex_site_ids, site_id_dimensions):
    """
    Create a Druid exclude filter for site dimension
    """
    dimension_filters = []
    and_filters = {}
    for dimension in site_id_dimensions:
        in_filters = {
            "type": "in",
            "dimension": dimension,
            "values": list(reindex_site_ids)
        }
        dimension_filters.append(in_filters)

    if len(dimension_filters) == 1:
        and_filters = dimension_filters[0]

    elif len(dimension_filters) > 1:
        and_filters = {
            "type": "and",
            "fields": dimension_filters
        }

    not_filter = {
        "type": "not",
        "field": and_filters
    }
    return not_filter


def get_reindex_site_ids(data_source, site_id_dimensions, args, ctx):
    """
    Get the intersection between sites from site-service and Druid
    """
    reindex_site_ids = {}
    offboarded_site_ids = get_offboarded_site_ids(args, ctx)
    logging.info("site ids from s2: %s", offboarded_site_ids)
    unique_site_ids = get_unique_site_ids_from_druid(
        data_source, site_id_dimensions, args, ctx)
    logging.info("site ids from druid: %s", unique_site_ids)
    reindex_site_ids = offboarded_site_ids.intersection(unique_site_ids)
    logging.info("site ids that need to offboard: %s", reindex_site_ids)
    return reindex_site_ids


def get_hourly_interval():
    """
    Get the Druid reindex interval for hourly tasks
    """
    start_time = (datetime.today() - timedelta(hours=24)).strftime('%Y-%m-%dT%H:00:00')
    end_time = (datetime.today() - timedelta(hours=2)).strftime('%Y-%m-%dT%H:00:00')
    time_interval = f"{start_time}/{end_time}"
    return time_interval


def get_daily_interval():
    """
    Get the Druid reindex interval for daily tasks
    """
    start_time = (datetime.today() - timedelta(days=7)).strftime('%Y-%m-%dT%H:00:00')
    end_time = (datetime.today() - timedelta(days=1)).strftime('%Y-%m-%dT%H:00:00')
    time_interval = f"{start_time}/{end_time}"
    return time_interval


def get_weekly_interval():
    """
    Get the Druid reindex interval for weekly tasks
    """
    start_time = '1970-01-01T00:00:00'
    end_time = (datetime.today() - timedelta(days=7)).strftime('%Y-%m-%dT%H:00:00')
    time_interval = f"{start_time}/{end_time}"
    return time_interval


def get_parser(args):
    """
    Parse the command
    """
    parser = argparse.ArgumentParser(description='Post Druid indexing tasks.')
    parser.add_argument('--data-source', '-d', type=str, default='correlated_flow_viz',
                        help='The name of Druid table (dataSource) to be rolled up')
    parser.add_argument('--url', '-u', metavar='url', type=str,
                        default='https://druid-coordinator:8281/',
                        help='Druid Overlord url')
    parser.add_argument('--coordinator-url', type=str, default='https://druid-coordinator:8281/',
                        help='Druid Coordinator url')
    parser.add_argument('--router-url', type=str, default='https://druid-router:8280/',
                        help='Druid router url')
    parser.add_argument('--broker-url', type=str, default='https://druid-broker:8282/',
                        help='Druid broker url')
    parser.add_argument('--s2-url', metavar='url', type=str, default='https://site-service:8710/',
                        help='amoeba url')
    parser.add_argument('--file', '-f', type=str,
                        required=True, help='Query JSON file')
    parser.add_argument('--submit-timeout', type=int, default=120,
                        help='Timeout (in seconds) for submitting tasks')
    parser.add_argument('--complete-timeout', type=int, default=1440000,
                        help='Timeout (in seconds) for completing tasks')
    parser.add_argument('--load-timeout', type=int, default=1440000,
                        help='Timeout (in seconds) for waiting for tasks to load')
    parser.add_argument('--quiet', '-q', action='store_true',
                        help='Suppress retryable errors')
    parser.add_argument('--user', type=str, default=None,
                        help='Basic auth username')
    parser.add_argument('--password', type=str,
                        default=None, help='Basic auth password')
    parser.add_argument('--granularity', '-g', type=str, required=True, default=None,
                        help='Segment granularity')
    parser.add_argument('--query-granularity', '-qg', type=str, required=False, default=None,
                        help='Query granularity')
    parser.add_argument('--num-days', '-n', type=int, default=None,
                        help='Number of days from when to rollup')
    parser.add_argument('--guard-interval-days', type=int, default=2,
                        help='Number of days back from now when rollup ends (default is 2 days)')
    parser.add_argument('--dimension', '-dm', type=str, nargs='+', required=True, default=None,
                        help='the dimensions of site id in druid table')
    parser.add_argument('--drop-existing', type=str,
                        default='true', help='Drop existing segments')
    parser.add_argument('--config-name', '-c', type=str, default='correlated_flow_viz_offboard_reindex',
                        help='config section name in file application.ini')
    return parser.parse_args(args)


def main():
    """
    Main function to run the reindex job
    """
    args = get_parser(sys.argv[1:])

    submit_timeout_at = time.time() + args.submit_timeout

    task_contents = read_json_file(args.file)
    task_json = json.loads(task_contents)

    datasource = args.data_source
    site_id_dimensions = args.dimension
    granularity = args.granularity
    config_section = args.config_name
    reindex_config = Config()
    logging.info("Beginning indexing data for %s", datasource)
    task_json["spec"]["ioConfig"]["inputSource"]["dataSource"] = datasource
    task_json["spec"]["dataSchema"]["dataSource"] = datasource
    task_json["context"]["druid.indexer.fork.property.druid.processing.buffer.sizeBytes"] = \
        reindex_config.get_config_value(config_section, PROCESSING_BUFFER_SIZE)
    task_json["context"]["druid.indexer.runner.javaOpts"] = \
        reindex_config.get_config_value(config_section, JAVA_OPTS)
    interval = ''
    if datasource.startswith('correlated_flow'):
        # for correlated_flow_viz, correlated_flow_rec
        # keep data within 2 hours
        # reindex data from past 24 hours to past 2 hours with granularity hour
        # reindex data from past 1 week to past 24 hours with granularity day
        # reindex data from 1900-01-01 to past 1 week with granularity week
        if granularity == 'hour':
            interval = get_hourly_interval()
        elif granularity == 'day':
            interval = get_daily_interval()
        elif granularity == 'week':
            interval = get_weekly_interval()
    else:
        interval = "1900-01-09T00:00:00.000Z/" + \
                   datetime.today().strftime('%Y-%m-%dT%H:00:00.000Z')
    logging.info("Druid reindex interval: %s", interval)
    task_json["spec"]["ioConfig"]["inputSource"]["interval"] = interval
    task_json["spec"]["ioConfig"]["dropExisting"] = args.drop_existing

    query_granularity = args.query_granularity
    if query_granularity is None:
        query_granularity = args.granularity
    task_json["spec"]["dataSchema"]["granularitySpec"]["queryGranularity"] = query_granularity
    task_json["spec"]["dataSchema"]["granularitySpec"]["segmentGranularity"] = args.granularity
    task_json["spec"]["dataSchema"]["granularitySpec"]["intervals"] = [interval]

    ctx = ssl.create_default_context()
    ctx.load_verify_locations("/etc/nsxi/certs/ca.crt")
    ctx.load_cert_chain(certfile="/etc/nsxi/certs/tls.crt",
                        keyfile="/etc/nsxi/certs/tls.key")

    reindex_site_ids = get_reindex_site_ids(
        datasource, site_id_dimensions, args, ctx)

    if len(reindex_site_ids) > 0:
        exclude_filter = get_exclude_filter(
            reindex_site_ids, site_id_dimensions)
        task_json["spec"]["ioConfig"]["inputSource"]["filter"] = exclude_filter
        task_contents = json.dumps(task_json)
        logging.info("Druid reindex spec: %s", task_contents)

        task_id = json.loads(
            post_task(args, task_contents, submit_timeout_at, ctx))["task"]
        logging.info("Druid reindex task id: %s", task_id)

    else:
        print("no need to reindex")


if __name__ == "__main__":
    main()
