"""
The module is for submitting Druid reindex hourly, daily and weekly job
"""
import argparse
import json
import logging
import os
import ssl
import sys
import time
from datetime import datetime, timedelta

from kubernetes import client, config

from tasks.config import Config
from tasks.remote_file_cleaner import get_druid_middlemanager_and_historical_pods, clean_stale_druid_task_files
from tasks.task_common import await_load_completion
from tasks.task_common import await_task_completion
from tasks.task_common import post_task
from tasks.task_common import read_json_file

FORMAT = '%(asctime)s %(levelname)s: %(message)s'
DATEFMT = '%Y-%m-%d %H:%M:%S'
PROCESSING_BUFFER_SIZE = 'PROCESSING_BUFFER_SIZE'
JAVA_OPTS = 'JAVA_OPTS'
MAX_ROWS_IN_MEMORY = 'MAX_ROWS_IN_MEMORY'
MAX_BYTES_IN_MEMORY = 'MAX_BYTES_IN_MEMORY'
MAX_ROWS_PER_SEGMENT = 'MAX_ROWS_PER_SEGMENT'
MAX_TOTAL_ROWS = 'MAX_TOTAL_ROWS'
MAX_COLUMNS_TO_MERGE = 'MAX_COLUMNS_TO_MERGE'
logging.basicConfig(format=FORMAT, datefmt=DATEFMT, level=logging.DEBUG)


def get_last_successful_job_timestamp(args):
    """
    Get last successful job timestamp
    """
    config.load_incluster_config()
    jobs = client.BatchV1Api().list_namespaced_job(os.getenv("POD_NAMESPACE"))
    datetimes = [job.status.completion_time for job in jobs.items
                 if job.metadata.name.startswith(args.job_type) and job.status.succeeded == 1]
    return sorted(datetimes)[-1]


def get_daily_rollup_interval(args):
    """
    Get Druid reindex daily/weekly task interval
    """
    # To preserve data with original granularity for the
    # past day end timestamp for rollup is 1 day before the
    # past day
    # for correlated_flow_viz, we always have the same granularity, no need to preserve
    # original granularity, we accept value passed by user from command line.
    interval_end_timestamp = datetime.today() - timedelta(days=args.guard_interval_days)
    # Read last timestamp when rollup succeeded from success file
    # If file does not exist or is corrupted reset the last success
    # timestamp to 30 days before current time.
    try:
        # remove timezone info, because datetime.today() doesn't generate timezone info
        # and timezone should be consistent in single kubernetes environment
        success_timestamp = get_last_successful_job_timestamp(
            args).replace(tzinfo=None)
    # pylint: disable=W0703
    except Exception as error:
        logging.info(
            "got exception when fetching last completed reindex job timestamp: %s", error)
        success_timestamp = interval_end_timestamp - timedelta(days=30)
    logging.info("last success timestamp: %s", success_timestamp)
    # this success_timestamp timestamp is "infinite" early for the first time or very early if a lot of failed task
    # for tasks all success case, we never reindexed interval of segments within the guard period since it was excluded
    # from end time of previous run
    last_guarded_timestamp = success_timestamp - timedelta(days=args.guard_interval_days)
    # ideally we would always want endtime - reindexPeriod as a start time.
    # However, if for some reason the task is running before the next cycle(retries or bugs), we would end up waste
    # resource on reindexing what was already indexed
    # we can take advantage of the recent guarded time which would be later than the normal fixed interval to reindex
    # less interval of data
    interval_start_timestamp = max(last_guarded_timestamp, interval_end_timestamp - timedelta(days=args.num_days))
    start_date = interval_start_timestamp.strftime('%Y-%m-%d')
    end_date = interval_end_timestamp.strftime('%Y-%m-%d')
    time_interval = f"{start_date}/{end_date}"
    return time_interval


def get_hourly_rollup_interval():
    """
    Get Druid reindex hourly task interval
    """
    interval_end_timestamp = datetime.today() - timedelta(hours=12)
    interval_start_timestamp = interval_end_timestamp - timedelta(hours=1)
    start_time = interval_start_timestamp.strftime('%Y-%m-%dT%H:00:00')
    end_time = interval_end_timestamp.strftime('%Y-%m-%dT%H:00:00')
    time_interval = f"{start_time}/{end_time}"
    return time_interval


def clean_stale_data(args):
    _, middlemanager_pods_set = get_druid_middlemanager_and_historical_pods()
    for middlemanager_pod in middlemanager_pods_set:
        logging.info("====================== cleaning for %s ======================",
                     middlemanager_pod)
        clean_stale_druid_task_files(middlemanager_pod, args)


def _submit_reindexing_task(args, retry):
    # clean previous stale data before each reindexing to avoid storage error
    clean_stale_data(args)

    submit_timeout_at = time.time() + args.submit_timeout
    complete_timeout_at = time.time() + args.complete_timeout

    task_contents = read_json_file(args.file)
    task_json = json.loads(task_contents)

    datasource = args.data_source
    config_section = args.config_name
    reindex_config = Config()

    logging.info("Beginning indexing data for %s ", datasource)
    task_json["context"]["druid.indexer.fork.property.druid.processing.buffer.sizeBytes"] = \
        reindex_config.get_config_value(config_section, PROCESSING_BUFFER_SIZE)
    task_json["context"]["druid.indexer.runner.javaOpts"] = \
        reindex_config.get_config_value(config_section, JAVA_OPTS)
    task_json["spec"]["tuningConfig"]["maxRowsInMemory"] = \
        str(int(reindex_config.get_config_value(config_section, MAX_ROWS_IN_MEMORY)) // retry)
    task_json["spec"]["tuningConfig"]["maxBytesInMemory"] = \
        str(int(reindex_config.get_config_value(config_section, MAX_BYTES_IN_MEMORY)) // retry)
    task_json["spec"]["tuningConfig"]["maxColumnsToMerge"] = \
        str(int(reindex_config.get_config_value(config_section, MAX_COLUMNS_TO_MERGE)) // retry)
    task_json["spec"]["tuningConfig"]["partitionsSpec"]["maxRowsPerSegment"] = \
        reindex_config.get_config_value(config_section, MAX_ROWS_PER_SEGMENT)
    task_json["spec"]["tuningConfig"]["partitionsSpec"]["maxTotalRows"] = \
        str(int(reindex_config.get_config_value(config_section, MAX_TOTAL_ROWS)) // retry)
    task_json["spec"]["ioConfig"]["inputSource"]["dataSource"] = datasource
    task_json["spec"]["dataSchema"]["dataSource"] = datasource

    if args.granularity == "hour":
        time_interval = get_hourly_rollup_interval()
    elif args.granularity == "day" or args.granularity == "week":
        time_interval = get_daily_rollup_interval(args)
    else:
        raise ValueError("The granularity " + args.granularity + "is not supported")

    task_json["spec"]["ioConfig"]["inputSource"]["interval"] = time_interval
    task_json["spec"]["dataSchema"]["granularitySpec"]["queryGranularity"] = args.granularity
    task_json["spec"]["dataSchema"]["granularitySpec"]["segmentGranularity"] = args.granularity
    task_contents = json.dumps(task_json)

    logging.info("task_spec for retry %s: %s", retry, task_contents)

    # SSL context initialization
    ctx = ssl.create_default_context()
    ctx.load_verify_locations("/etc/nsxi/certs/ca.crt")

    task_id = json.loads(post_task(args, task_contents,
                                   submit_timeout_at, ctx))["task"]

    # pylint: disable=W1201,C0209
    logging.info('\033[1m' + "Task started: " +
                 '\033[0m' + "{0}".format(task_id))
    logging.info('\033[1m' + "Task log:     " + '\033[0m' +
                 "{0}/druid/indexer/v1/task/{1}/log".format(args.url.rstrip("/"), task_id))
    logging.info('\033[1m' + "Task status:  " + '\033[0m' +
                 "{0}/druid/indexer/v1/task/{1}/status".format(args.url.rstrip("/"), task_id))

    task_status = await_task_completion(
        args, task_id, complete_timeout_at, ctx)

    logging.info("Task finished with status: %s", task_status)
    if task_status == 'SUCCESS':
        logging.info(
            "Completed indexing data for %s. Now loading indexed data onto the cluster...", datasource)
        load_timeout_at = time.time() + args.load_timeout
        await_load_completion(args, datasource, load_timeout_at, ctx)
        sys.exit(0)
    else:
        logging.info("Reindexing task failed, retry time: %s", retry)


def main():
    """
    Main function to run a Druid reindex job
    """
    parser = argparse.ArgumentParser(description='Post Druid indexing tasks.')
    parser.add_argument('--data-source', '-d', type=str, required=True,
                        help='The name of Druid table (dataSource) to be rolled up')
    parser.add_argument('--url', '-u', metavar='url', type=str,
                        default='https://druid-coordinator:8281/', help='Druid Overlord url')
    parser.add_argument('--coordinator-url', type=str,
                        default='https://druid-coordinator:8281/', help='Druid Coordinator url')
    parser.add_argument('--file', '-f', type=str,
                        required=True, help='Query JSON file')
    parser.add_argument('--submit-timeout', type=int, default=120,
                        help='Timeout (in seconds) for submitting tasks')
    parser.add_argument('--complete-timeout', type=int, default=72000,
                        help='Timeout (in seconds) for completing tasks')
    parser.add_argument('--load-timeout', type=int, default=14400,
                        help='Timeout (in seconds) for waiting for tasks to load')
    parser.add_argument('--quiet', '-q', action='store_true',
                        help='Suppress retryable errors')
    parser.add_argument('--user', type=str, default=None,
                        help='Basic auth username')
    parser.add_argument('--password', type=str,
                        default=None, help='Basic auth password')
    parser.add_argument('--granularity', '-g', type=str, required=True,
                        default=None, help='Segment and query granularity')
    parser.add_argument('--num-days', '-n', type=int, required=True,
                        default=None, help='Number of days from when to rollup')
    parser.add_argument('--guard-interval-days', type=int, default=2,
                        help='Number of days back from now when rollup ends (default is 2 days)')
    parser.add_argument('--job-type', type=str, required=True,
                        help='reindex job type, this should the name of the cronjob')
    parser.add_argument('--config-name', '-c', type=str, default='reindex',
                        help='config section name in file application.ini')
    parser.add_argument('--dry', action="store_true", default=False,
                        help='--dry to dry run, no real file deletion for stale data')
    parser.add_argument('--retry', type=int, default=2,
                        help='Number of retries for reindexing task, each time the reindexing parameters are turned '
                             'down')
    args = parser.parse_args()
    for i in range(1, args.retry + 1):
        _submit_reindexing_task(args, i)
    logging.info('Exhausted all %s reties', args.retry)
    sys.exit(0)


if __name__ == "__main__":
    main()
