"""
The module to clean tmp file in Druid
"""
import argparse
import json
import logging
import os
import subprocess
from datetime import datetime, timedelta

JVM_TEMP_DIR = "/data/tmp"
DRUID_TASK_DIR = "/data/druid/current-task"

DEFAULT_JVM_TEMP_OURGE_AGE = 86400  # 24h
FORMAT = '%(asctime)s %(levelname)s: %(message)s'
DATEFMT = '%Y-%m-%d %H:%M:%S'
logging.basicConfig(format=FORMAT, datefmt=DATEFMT, level=logging.DEBUG)


# execute command on pod via kubectl exec -it
# if execution is not correctly executed, will return None
def execute_command_for_pod(pod, command):
    """
    Execute command for pod
    """
    cmd = f"kubectl exec -i {pod} -c druid -- {command}"
    try:
        output = subprocess.check_output(cmd, shell=True)
        return output
    except subprocess.CalledProcessError:
        logging.info("execution error for command: %s", cmd)
        return None


def rm_rf(pod, path):
    """
    Execute command rm -rf on pod
    """
    command = f"rm -rf {path}"
    return execute_command_for_pod(pod, command)


# list file/folder names in the dir_path on pod, with last modified time
# return value is a dict where key is filename, value is datetime object of last modified time
def listdir_and_modified_time_for_pod(pod, dir_path):
    """
    list file/folder names in the dir_path on pod, with last modified time
    Return value is a dict where key is filename, value is datetime object of last modified time
    """
    command = f"ls -l --time-style=long-iso {dir_path}"
    file_modified_time = {}

    raw_output = execute_command_for_pod(pod, command)
    if raw_output is None:
        logging.info("failed to listdir for pod %s", pod)
        return file_modified_time

    lines = raw_output.decode().splitlines()

    for line in lines:
        ls_columns = line.split()
        if len(ls_columns) < 8:
            continue  # not a file entry line

        date_s, time_s, name = ls_columns[5:8]
        date = datetime.fromisoformat(" ".join([date_s, time_s]))
        file_modified_time[name] = date

    return file_modified_time


# get druid historical pod names and middlemanager names


def get_druid_middlemanager_and_historical_pods():
    """
    Get druid historical pod names and middle manager names
    """
    cmd = "kubectl get pod -o name | grep druid"
    historical_pods, middlemanager_pods = set(), set()
    try:
        raw_output = subprocess.check_output(cmd, shell=True)
        pods = raw_output.decode().splitlines()
        for pod in pods:
            if pod.startswith("pod/druid-historical") \
                    or pod.startswith("pod/druid-config-historical"):
                historical_pods.add(pod[4:])
            elif pod.startswith("pod/druid-middle-manager"):
                middlemanager_pods.add(pod[4:])
    except subprocess.CalledProcessError:
        logging.info("execution error for command: %s", cmd)
    return historical_pods, middlemanager_pods


def get_running_task_for_pod(pod):
    """
    Read restore.json file to get running tasks
    """
    restore_file_path = os.path.join(DRUID_TASK_DIR, "restore.json")
    command = f"cat {restore_file_path}"

    raw_output = execute_command_for_pod(pod, command)
    if raw_output is None:
        logging.info("no restore.json in pod %s", pod)
        return set()
    restore_object = json.loads(raw_output.decode())
    return set(restore_object["runningTasks"])


def clean_jvm_temp_files(pod, args):
    """
    Cleans jvm io temp file older than jvm_temp_purge_age
    """
    timedelta_threshold = timedelta(seconds=args.age)
    for file_name, modified_time in listdir_and_modified_time_for_pod(pod, JVM_TEMP_DIR).items():
        if datetime.now() - modified_time > timedelta_threshold:

            aged_tmp_file_path = os.path.join(JVM_TEMP_DIR, file_name)
            if not args.dry:
                rm_rf(pod, aged_tmp_file_path)
            logging.info('cleaned tmp file %s last modified %s', aged_tmp_file_path, modified_time)


def clean_stale_druid_task_files(pod, args):
    """
    Clean file for stale tasks
    """
    running_tasks = get_running_task_for_pod(pod)

    logging.info('running tasks %s', running_tasks)

    for file_name, _ in listdir_and_modified_time_for_pod(pod, DRUID_TASK_DIR).items():

        # skip druid internally used file/folders
        if file_name in ["completedTasks", "assignedTasks",
                         "restore.json", "workerTaskManagerTmp"]:
            continue

        task_dict = listdir_and_modified_time_for_pod(pod, os.path.join(DRUID_TASK_DIR, file_name))

        # skip empty slot
        if len(task_dict) == 0:
            logging.info('skipping empty folder %s', file_name)
            continue

        # skip running task
        skip_file = False
        for task_name, _ in task_dict.items():
            if task_name in running_tasks:
                logging.info('task %s is still running, skipping folder %s', task_name, file_name)
                skip_file = True
                break
        if skip_file:
            continue

        stale_task_path = os.path.join(DRUID_TASK_DIR, file_name)
        if not args.dry:
            rm_rf(pod, stale_task_path)
        logging.info('cleaned stale task folder %s', stale_task_path)


def main():
    """
    Main function
    """
    parser = argparse.ArgumentParser(description='Druid cleaner job.')
    parser.add_argument('--age', '-a', type=int, default=DEFAULT_JVM_TEMP_OURGE_AGE,
                        help='the age in seconds, '
                             'jvm io temp file/folder older than this age will be cleaned')
    parser.add_argument('--dry', action="store_true", default=False,
                        help='--dry to dry run, no real file deletion')
    args = parser.parse_args()

    historical_pods_set, middlemanager_pods_set = get_druid_middlemanager_and_historical_pods()

    for historical_pod in historical_pods_set:
        logging.info("====================== cleaning for %s ======================",
                     historical_pod)
        clean_jvm_temp_files(historical_pod, args)

    for middlemanager_pod in middlemanager_pods_set:
        logging.info("====================== cleaning for %s ======================",
                     middlemanager_pod)
        clean_jvm_temp_files(middlemanager_pod, args)
        clean_stale_druid_task_files(middlemanager_pod, args)


if __name__ == "__main__":
    main()
