# pylint: disable=duplicate-code
"""
The module is common util for druid tasks
"""
import base64
import json
import logging
import re
import sys
import time
from json import JSONDecodeError
from urllib.error import HTTPError
from urllib.error import URLError
from urllib.parse import urlparse
from urllib.request import Request
from urllib.request import quote
from urllib.request import urlopen

FORMAT = '%(asctime)s %(levelname)s: %(message)s'
DATEFMT = '%Y-%m-%d %H:%M:%S'
logging.basicConfig(format=FORMAT, datefmt=DATEFMT, level=logging.DEBUG)


def read_json_file(file_path):
    """
    Read file
    """
    with open(file_path, 'r', encoding="utf-8") as file:
        contents = file.read()
        # We don't use the parsed data, but we want to throw early if it's invalid
        try:
            json.loads(contents)
        except JSONDecodeError as ex:
            logging.info('Invalid JSON in task file "%s": %s',
                         file_path, repr(ex))
            sys.exit(0)
        return contents


def add_basic_auth_header(args, req):
    """
    Add basic auth header
    """
    if args.user is not None:
        basic_auth_encoded = base64.b64encode(f"{args.user}:{args.password}")
        req.add_header("Authorization", f"Basic {basic_auth_encoded}")


def terminate_task(args, task_id, ctx=None):
    logging.info("trying to terminate the taskid: " + task_id)
    try:
        url = args.url.rstrip("/") + "/druid/indexer/v1/task/" + task_id + "/shutdown"
        req = Request(url, method="POST", headers={'Content-Type': 'application/json'})
        add_basic_auth_header(args, req)
        with urlopen(req, None, 10, context=ctx) as response:
            if response.status != 200:
                logging.warning("response status is: " + str(response.status))
                sys.exit(0)
    except Exception as e:
        logging.error("Error while trying to terminate the task: " + task_id + ": " + str(e))
        sys.exit(0)


# Keep trying until timeout_at, then abandon further retries
def post_task(args, task_json, timeout_at, ctx=None):
    """
    Post task
    """
    try:
        url = args.url.rstrip("/") + "/druid/indexer/v1/task"
        data = task_json.encode("utf-8")
        req = Request(url, None, {'Content-Type': 'application/json'})
        add_basic_auth_header(args, req)
        timeleft = timeout_at - time.time()
        response_timeout = min(max(timeleft, 5), 10)
        with urlopen(req, data, response_timeout, context=ctx) as response:
            return response.read().rstrip()
    except URLError as error:
        # pylint: disable=E1101
        if isinstance(error, HTTPError) and error.code >= 400 and error.code <= 500:
            # 4xx (problem with the request) or 500 (something wrong on the server)
            raise_friendly_error(error)
        elif time.time() >= timeout_at:
            # No further retries
            raise_friendly_error(error)
        elif isinstance(error, HTTPError) and error.code in [301, 302, 303, 305, 307] and \
                error.getheader("Location") is not None:
            # Set the new location in args.url,
            # so it can be used by await_task_completion and re-issue the request
            location = urlparse(error.getheader("Location"))
            args.url = f"{location.scheme}://{location.netloc}"
            logging.info(
                "Redirect response received, setting url to [%s]", args.url)
            return post_task(args, task_json, timeout_at, ctx)
        else:
            # If at first you don't succeed, try, try again!
            sleep_time = 5
            if not args.quiet:
                extra = ''
                if hasattr(error, 'read'):
                    extra = error.read().rstrip()
                logging.info("Waiting up to %s seconds for indexing service [%s] to become "
                             "available. [Got: %s %s]",
                             max(sleep_time, int(timeout_at - time.time())),
                             args.url,
                             str(error),
                             extra)
            time.sleep(sleep_time)
            return post_task(args, task_json, timeout_at, ctx)
        return None


# Keep trying until timeout_at, then abandon further retries
def get_s2_sites(args, timeout_at, ctx=None):
    """
    Get site info for site service
    """
    try:
        url = args.s2_url.rstrip("/") + "/api/v1/intelligence/site-service/sites"
        req = Request(url, None, {'Accept': '*/*'})
        add_basic_auth_header(args, req)
        timeleft = timeout_at - time.time()
        response_timeout = min(max(timeleft, 5), 10)
        with urlopen(req, None, response_timeout, context=ctx) as response:
            return response.read().rstrip()
    except URLError as error:
        # pylint: disable=E1101
        if isinstance(error, HTTPError) and error.code >= 400 and error.code <= 500:
            # 4xx (problem with the request) or 500 (something wrong on the server)
            raise_friendly_error(error)
        elif time.time() >= timeout_at:
            # No further retries
            raise_friendly_error(error)
        elif isinstance(error, HTTPError) and error.code in [301, 302, 303, 305, 307] and \
                error.getheader("Location") is not None:
            # Set the new location in args.url,
            # so it can be used by await_task_completion and re-issue the request
            location = urlparse(error.getheader("Location"))
            args.url = f"{location.scheme}://{location.netloc}"
            logging.info(
                "Redirect response received, setting url to [%s]", url)
            return get_s2_sites(args, timeout_at, ctx)
        else:
            # If at first you don't succeed, try, try again!
            sleep_time = 5
            time.sleep(sleep_time)
            return get_s2_sites(args, timeout_at, ctx)
        return None


# Keep trying until timeout_at, maybe die then
# pylint: disable=R1710
def post_query(args, task_json, timeout_at, ctx=None):
    """
    Post Query to druid
    """
    try:
        url = args.broker_url.rstrip("/") + "/druid/v2/?pretty"
        data = task_json.encode("utf-8")
        req = Request(url, None, {'Content-Type': 'application/json'})
        add_basic_auth_header(args, req)
        timeleft = timeout_at - time.time()
        response_timeout = min(max(timeleft, 5), 10)
        with urlopen(req, data, response_timeout, context=ctx) as response:
            return response.read().rstrip()
    except URLError as error:
        # pylint: disable=E1101
        if isinstance(error, HTTPError) and error.code >= 400 and error.code <= 500:
            # 4xx (problem with the request) or 500 (something wrong on the server)
            raise_friendly_error(error)
        elif time.time() >= timeout_at:
            # No further retries
            raise_friendly_error(error)
        elif isinstance(error, HTTPError) and error.code in [301, 302, 303, 305, 307] and \
                error.getheader("Location") is not None:
            # Set the new location in args.url,
            # so it can be used by await_task_completion and re-issue the request
            location = urlparse(error.getheader("Location"))
            args.url = f"{location.scheme}://{location.netloc}"
            logging.info(
                "Redirect response received, setting url to [%s]", args.broker_url)
            return post_query(args, task_json, timeout_at, ctx)
        else:
            # If at first you don't succeed, try, try again!
            sleep_time = 5
            if not args.quiet:
                extra = ''
                if hasattr(error, 'read'):
                    extra = error.read().rstrip()
                logging.info("Waiting up to %s seconds for indexing service [%s] to "
                             "become available. [Got: %s %s]",
                             max(sleep_time, int(timeout_at - time.time())),
                             args.broker_url,
                             str(error),
                             extra)
            time.sleep(sleep_time)
            return post_query(args, task_json, timeout_at, ctx)


# Keep trying until timeout_at, maybe die then
def await_task_completion(args, task_id, timeout_at, ctx=None):
    """
    Await Druid task completion
    """
    while True:
        url = args.url.rstrip("/") + f"/druid/indexer/v1/task/{quote(task_id)}/status"
        req = Request(url)
        add_basic_auth_header(args, req)
        timeleft = timeout_at - time.time()
        response_timeout = min(max(timeleft, 5), 10)
        with urlopen(req, None, response_timeout, context=ctx) as response:
            response_obj = json.loads(response.read())
            response_status_code = response_obj["status"]["status"]
            if response_status_code in ['SUCCESS', 'FAILED']:
                return response_status_code

            if time.time() < timeout_at:
                if not args.quiet:
                    logging.info("Task %s still running...", task_id)
                timeleft = timeout_at - time.time()
                time.sleep(min(30, timeleft))
            else:
                terminate_task(args, task_id, ctx)
                logging.error(f"Task {task_id} did not finish in time!")
                sys.exit(0)


def raise_friendly_error(error):
    """
    Raise friendly error
    """
    if isinstance(error, HTTPError):
        text = error.read().decode('utf-8').strip()
        result = re.search(r"<pre>(.*?)</pre>", text, re.DOTALL)
        if result:
            text = result.group(1).strip()
        raise Exception(f"HTTP Error {error.code}: {error.reason}, check overlord log "
                        f"for more details.\n{text}")
    raise error


def await_load_completion(args, datasource, timeout_at, ctx=None):
    """
    Await load status completion
    """
    while True:
        url = args.coordinator_url.rstrip(
            "/") + "/druid/coordinator/v1/loadstatus"
        req = Request(url)
        add_basic_auth_header(args, req)
        timeleft = timeout_at - time.time()
        response_timeout = min(max(timeleft, 5), 10)
        with urlopen(req, None, response_timeout, context=ctx) as response:
            response_obj = json.loads(response.read())
            load_status = response_obj.get(datasource, 0.0)
            if load_status >= 100.0:
                logging.info(
                    "%s loading complete! You may now query your data", datasource)
                return

            if time.time() < timeout_at:
                if not args.quiet:
                    logging.info("%s is %s %% finished loading...",
                                 datasource, load_status)
                timeleft = timeout_at - time.time()
                time.sleep(min(5, timeleft))
            else:
                raise Exception(
                    f"{datasource} was not loaded in time!")
